/*  mkBuffer -- Create a shared memory bounded buffer.
    Copyright (C) 2012 Lawrence Sebald

    Permission to use, copy, modify, and/or distribute this software for any
    purpose with or without fee is hereby granted, provided that the above
    copyright notice and this permission notice appear in all copies.

    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
    WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
    SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
    WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
    OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
    CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <semaphore.h>
#include <sys/mman.h>

#include "buffer.h"

int main(int argc, char *argv[]) {
    int fd, bufsz;
    buffer_t *buf;
    size_t totalsz;
    sem_t *sem;

    if(argc != 2 || (bufsz = atoi(argv[1])) <= 0) {
        printf("Usage: %s N\nWhere N is the size of the buffer to create.\n",
               argv[0]);
        exit(EXIT_FAILURE);
    }

    totalsz = sizeof(buffer_t) + bufsz * sizeof(int);

    /* Create the semaphore we need */
    sem = sem_open(SEM_NAME, O_CREAT | O_EXCL, 0600);
    if(sem == SEM_FAILED) {
        perror("sem_open");
        exit(EXIT_FAILURE);
    }

    /* Lock the semaphore before we do anything else */
    if(sem_wait(sem)) {
        perror("sem_wait");
        goto err_sem;
    }

    /* Create the shared memory region */
    fd = shm_open(SHM_REGION_NAME, O_CREAT | O_RDWR, S_IRUSR | S_IWUSR);
    if(fd < 0) {
        perror("shm_open");
        goto err_semp;
    }

    /* Set its size to an appropriate value */
    if(ftruncate(fd, totalsz) < 0) {
        perror("ftruncate");
        goto err_out;
    }

    /* mmap the region so we can set everything up nicely */
    buf = (buffer_t *)mmap(NULL, totalsz, PROT_READ | PROT_WRITE, MAP_SHARED,
                           fd, 0);
    if(buf == MAP_FAILED) {
        perror("mmap");
        goto err_out;
    }

    /* Set up the variables in the shared memory */
    buf->max_items = bufsz;
    buf->cur_items = 0;

    munmap(buf, totalsz);
    close(fd);

    sem_post(sem);
    sem_close(sem);

    return 0;

err_out:
    close(fd);
    shm_unlink(SHM_REGION_NAME);
err_semp:
    sem_post(sem);
err_sem:
    sem_close(sem);
    sem_unlink(SEM_NAME);
    exit(EXIT_FAILURE);
}
