/*  putBuffer -- Put an item into a shared memory bounded buffer.
    Copyright (C) 2012 Lawrence Sebald

    Permission to use, copy, modify, and/or distribute this software for any
    purpose with or without fee is hereby granted, provided that the above
    copyright notice and this permission notice appear in all copies.

    THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
    WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
    MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
    SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
    WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION
    OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN
    CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <semaphore.h>
#include <sys/mman.h>

#include "buffer.h"

int main(int argc, char *argv[]) {
    sem_t *sem;
    int fd;
    buffer_t *buf;
    size_t totalsz;
    int rv = 0, item;

    if(argc != 2) {
        printf("Usage: %s N\nWhere N is the item to add to the buffer.\n",
               argv[0]);
        exit(EXIT_FAILURE);
    }

    item = atoi(argv[1]);

    /* Open and lock the semaphore */
    sem = sem_open(SEM_NAME, 0, 0);
    if(sem == SEM_FAILED) {
        perror("sem_open");
        exit(EXIT_FAILURE);
    }

retry:
    if(sem_wait(sem)) {
        perror("sem_wait");
        rv = -1;
        goto err_sem;
    }

    /* Open the shared memory region */
    fd = shm_open(SHM_REGION_NAME, O_RDWR, 0);
    if(fd < 0) {
        perror("shm_open");
        rv = -1;
        goto err_semp;
    }

    /* mmap the header of the region so we can read the real size and the number
       of items in the buffer... */
    buf = (buffer_t *)mmap(NULL, sizeof(buffer_t), PROT_READ, MAP_SHARED, fd,
                           0);
    if(buf == MAP_FAILED) {
        perror("mmap");
        rv = -1;
        goto err_out;
    }

    /* If the buffer is full, we need to wait... Not exactly the most beautiful
       way to do this, but it works and its portable... */
    if(buf->max_items == buf->cur_items) {
        munmap(buf, sizeof(buffer_t));
        close(fd);
        sem_post(sem);
        sleep(1);
        goto retry;
    }

    totalsz = sizeof(buffer_t) + buf->max_items * sizeof(int);
    munmap(buf, sizeof(buffer_t));

    /* Re-mmap now that we know the size */
    buf = (buffer_t *)mmap(NULL, totalsz, PROT_READ | PROT_WRITE, MAP_SHARED,
                           fd, 0);
    if(buf == MAP_FAILED) {
        perror("mmap");
        rv = -1;
        goto err_out;
    }

    buf->items[buf->cur_items++] = item;

err_out:
    munmap(buf, totalsz);
    close(fd);
err_semp:
    sem_post(sem);
err_sem:
    sem_close(sem);

    return rv;
}
